'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _nullthrows = require('nullthrows');

var _nullthrows2 = _interopRequireDefault(_nullthrows);

var _pcap = require('pcap');

var _pcap2 = _interopRequireDefault(_pcap);

var _MacAddresses = require('./MacAddresses');

var _MacAddresses2 = _interopRequireDefault(_MacAddresses);

var _NetworkInterfaces = require('./NetworkInterfaces');

var _NetworkInterfaces2 = _interopRequireDefault(_NetworkInterfaces);

var _Packets = require('./Packets');

var _Packets2 = _interopRequireDefault(_Packets);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

let pcapSession;

function getPcapSession(interfaceName) {
  if (!pcapSession) {
    pcapSession = _Packets2.default.createCaptureSession(interfaceName);
  } else {
    _assert2.default.equal(interfaceName, pcapSession.device_name, 'The existing pcap session must be listening on the specified interface');
  }
  return pcapSession;
}

class DashButton {

  constructor(macAddress, options = {}) {
    this._macAddress = macAddress.toLowerCase();
    this._networkInterface = options.networkInterface || (0, _nullthrows2.default)(_NetworkInterfaces2.default.getDefault());
    this._packetListener = this._handlePacket.bind(this);
    this._dashListeners = new Set();
    this._isResponding = false;
  }

  addListener(listener) {
    if (!this._dashListeners.size) {
      let session = getPcapSession(this._networkInterface);
      session.addListener('packet', this._packetListener);
    }

    // We run the listeners with Promise.all, which rejects early as soon as
    // any of its promises are rejected. Since we want to wait for all of the
    // listeners to finish we need to catch any errors they may throw.
    let guardedListener = this._createGuardedListener(listener);
    this._dashListeners.add(guardedListener);

    return new Subscription(() => {
      this._dashListeners.delete(guardedListener);
      if (!this._dashListeners.size) {
        let session = getPcapSession(this._networkInterface);
        session.removeListener('packet', this._packetListener);
        if (!session.listenerCount('packet')) {
          session.close();
        }
      }
    });
  }

  _createGuardedListener(listener) {
    return async (...args) => {
      try {
        await listener(...args);
      } catch (error) {
        return error;
      }
    };
  }

  async _handlePacket(rawPacket) {
    if (this._isResponding) {
      return;
    }

    let packet = _pcap2.default.decode(rawPacket);
    let macAddress = _MacAddresses2.default.getEthernetSource(packet);
    if (macAddress !== this._macAddress) {
      return;
    }

    this._isResponding = true;
    try {
      // The listeners are guarded so this should never throw, but wrap it in
      // try-catch to be defensive
      let listeners = Array.from(this._dashListeners);
      let errors = await Promise.all(listeners.map(listener => listener(packet)));
      for (let error of errors) {
        if (error) {
          console.error(`Listener threw an uncaught error:\n${error.stack}`);
        }
      }
    } finally {
      this._isResponding = false;
    }
  }
}

exports.default = DashButton;
class Subscription {

  constructor(onRemove) {
    this._remove = onRemove;
  }

  remove() {
    if (!this._remove) {
      return;
    }
    this._remove();
    delete this._remove;
  }
}
module.exports = exports['default'];